from unittest import TestCase

import orjson as json

from panther import Panther, status
from panther.authentications import QueryParamJWTAuthentication
from panther.configs import config
from panther.test import WebsocketClient
from panther.websocket import GenericWebsocket


class WithoutAcceptWebsocket(GenericWebsocket):
    async def connect(self):
        pass


class CloseOnConnectWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.close()


class CustomCloseWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.close(code=status.WS_1013_TRY_AGAIN_LATER, reason='Come Back Later')


class MessageOnConnectWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send('Hello')
        await self.close()


class QueryParamWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send(self.query_params)
        await self.close()


class PathVariableWebsocket(GenericWebsocket):
    async def connect(self, name: str, age: int, is_male: bool):
        await self.accept()
        await self.send(
            f'{type(name).__name__}({name}), {type(age).__name__}({age}), {type(is_male).__name__}({is_male})',
        )
        await self.close()


class SendAllTypesWebsocket(GenericWebsocket):
    async def connect(self):
        await self.accept()
        # String
        await self.send('Hello Again')
        # Int
        await self.send(12)
        # Dict
        await self.send({'detail': 'ok'})
        # List
        await self.send([1, 2, 3, 'Ali', 4])
        # Tuple
        await self.send((1, 2, 3, 'Ali', 4))
        # Bytes
        await self.send(b'It Is Value Of A File')
        # Nothing
        await self.send()
        # None
        await self.send(None)
        # Close
        await self.close()


class WebsocketWithoutAuthentication(GenericWebsocket):
    async def connect(self):
        await self.accept()
        await self.send(str(self.user))
        await self.close()


class WebsocketWithAuthentication(GenericWebsocket):
    auth = QueryParamJWTAuthentication

    async def connect(self):
        await self.accept()
        await self.send(self.user)
        await self.close()


class Permission:
    async def __call__(self, connection) -> bool:
        return connection.path == '/with-permission'


class WebsocketWithPermission(GenericWebsocket):
    permissions = [Permission]

    async def connect(self):
        await self.accept()
        await self.send('ok')
        await self.close()


class WebsocketWithoutPermission(GenericWebsocket):
    permissions = [Permission]

    async def connect(self):
        await self.accept()
        await self.send('no')
        await self.close()


urls = {
    'without-accept': WithoutAcceptWebsocket,
    'close-on-connect': CloseOnConnectWebsocket,
    'custom-close': CustomCloseWebsocket,
    'message-after-connect': MessageOnConnectWebsocket,
    'query-params': QueryParamWebsocket,
    'path-variable/<name>/<age>/<is_male>/': PathVariableWebsocket,
    'all-types': SendAllTypesWebsocket,
    'without-auth': WebsocketWithoutAuthentication,
    'with-auth': WebsocketWithAuthentication,
    'with-permission': WebsocketWithPermission,
    'without-permission': WebsocketWithoutPermission,
}


class TestWebsocket(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.app = Panther(__name__, configs=__name__, urls=urls)

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    def test_without_accept(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-accept')
        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_close_on_connect(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('close-on-connect')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.close'
        assert responses[1]['code'] == 1000
        assert responses[1]['reason'] == ''

    def test_custom_close(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('custom-close')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.close'
        assert responses[1]['code'] == 1013
        assert responses[1]['reason'] == 'Come Back Later'

    def test_query_params(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('query-params', query_params={'name': 'ali', 'age': 27})
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        params = json.loads(responses[1]['text'])
        assert [*params.keys()] == ['name', 'age']
        assert params['name'] == 'ali'
        assert params['age'] == '27'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_message_after_connect(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('message-after-connect')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'Hello'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_path_variables(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('path-variable/Ali/25/true')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'str(Ali), int(25), bool(True)'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_all_types(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('all-types')

        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        # String
        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'Hello Again'
        assert 'bytes' not in responses[1]
        # Int
        assert responses[2]['type'] == 'websocket.send'
        assert responses[2]['text'] == '12'
        assert 'bytes' not in responses[2]
        # Dict
        assert responses[3]['type'] == 'websocket.send'
        assert responses[3]['text'] == '{"detail":"ok"}'
        assert 'bytes' not in responses[3]
        # List
        assert responses[4]['type'] == 'websocket.send'
        assert responses[4]['text'] == '[1,2,3,"Ali",4]'
        assert 'bytes' not in responses[4]
        # Tuple
        assert responses[5]['type'] == 'websocket.send'
        assert responses[5]['text'] == '[1,2,3,"Ali",4]'
        assert 'bytes' not in responses[5]
        # Bytes
        assert responses[6]['type'] == 'websocket.send'
        assert responses[6]['bytes'] == b'It Is Value Of A File'
        assert 'text' not in responses[6]
        # Nothing
        assert responses[7]['type'] == 'websocket.send'
        assert responses[7]['text'] == 'null'
        assert 'bytes' not in responses[7]
        # None
        assert responses[8]['type'] == 'websocket.send'
        assert responses[8]['text'] == 'null'
        assert 'bytes' not in responses[8]

        # Close
        assert responses[9]['type'] == 'websocket.close'
        assert responses[9]['code'] == 1000
        assert responses[9]['reason'] == ''

    def test_without_auth(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-auth')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'None'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_with_auth_failed(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('with-auth')

        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'null'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_with_auth_success(self):
        global WS_AUTHENTICATION, SECRET_KEY, DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.PantherDBConnection',
            },
        }
        WS_AUTHENTICATION = 'panther.authentications.QueryParamJWTAuthentication'
        SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
        token = 'eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxfQ.AF3nsj8IQ6t0ncqIx4quoyPfYaZ-pqUOW4z_euUztPM'
        app = Panther(__name__, configs=__name__, urls=urls)

        ws = WebsocketClient(app=app)
        with self.assertLogs(level='ERROR') as captured:
            responses = ws.connect('with-auth', query_params={'authorization': token})

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'QueryParamJWTAuthentication Error: "User not found"'

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''

    def test_with_permission(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('with-permission')

        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == 'ok'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''

    def test_without_permission(self):
        ws = WebsocketClient(app=self.app)
        responses = ws.connect('without-permission')

        assert responses[0]['type'] == 'websocket.close'
        assert responses[0]['code'] == 1000
        assert responses[0]['reason'] == ''



"""
HTTP codes
https://www.iana.org/assignments/http-status-codes/http-status-codes.xhtml
And RFC 2324 - https://tools.ietf.org/html/rfc2324

WEBSOCKET codes
https://www.iana.org/assignments/websocket/websocket.xhtml
"""

HTTP_100_CONTINUE = 100
HTTP_101_SWITCHING_PROTOCOLS = 101
HTTP_102_PROCESSING = 102
HTTP_103_EARLY_HINTS = 103
HTTP_200_OK = 200
HTTP_201_CREATED = 201
HTTP_202_ACCEPTED = 202
HTTP_203_NON_AUTHORITATIVE_INFORMATION = 203
HTTP_204_NO_CONTENT = 204
HTTP_205_RESET_CONTENT = 205
HTTP_206_PARTIAL_CONTENT = 206
HTTP_207_MULTI_STATUS = 207
HTTP_208_ALREADY_REPORTED = 208
HTTP_226_IM_USED = 226
HTTP_300_MULTIPLE_CHOICES = 300
HTTP_301_MOVED_PERMANENTLY = 301
HTTP_302_FOUND = 302
HTTP_303_SEE_OTHER = 303
HTTP_304_NOT_MODIFIED = 304
HTTP_305_USE_PROXY = 305
HTTP_306_RESERVED = 306
HTTP_307_TEMPORARY_REDIRECT = 307
HTTP_308_PERMANENT_REDIRECT = 308
HTTP_400_BAD_REQUEST = 400
HTTP_401_UNAUTHORIZED = 401
HTTP_402_PAYMENT_REQUIRED = 402
HTTP_403_FORBIDDEN = 403
HTTP_404_NOT_FOUND = 404
HTTP_405_METHOD_NOT_ALLOWED = 405
HTTP_406_NOT_ACCEPTABLE = 406
HTTP_407_PROXY_AUTHENTICATION_REQUIRED = 407
HTTP_408_REQUEST_TIMEOUT = 408
HTTP_409_CONFLICT = 409
HTTP_410_GONE = 410
HTTP_411_LENGTH_REQUIRED = 411
HTTP_412_PRECONDITION_FAILED = 412
HTTP_413_REQUEST_ENTITY_TOO_LARGE = 413
HTTP_414_REQUEST_URI_TOO_LONG = 414
HTTP_415_UNSUPPORTED_MEDIA_TYPE = 415
HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE = 416
HTTP_417_EXPECTATION_FAILED = 417
HTTP_418_IM_A_TEAPOT = 418
HTTP_421_MISDIRECTED_REQUEST = 421
HTTP_422_UNPROCESSABLE_ENTITY = 422
HTTP_423_LOCKED = 423
HTTP_424_FAILED_DEPENDENCY = 424
HTTP_425_TOO_EARLY = 425
HTTP_426_UPGRADE_REQUIRED = 426
HTTP_428_PRECONDITION_REQUIRED = 428
HTTP_429_TOO_MANY_REQUESTS = 429
HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE = 431
HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS = 451
HTTP_500_INTERNAL_SERVER_ERROR = 500
HTTP_501_NOT_IMPLEMENTED = 501
HTTP_502_BAD_GATEWAY = 502
HTTP_503_SERVICE_UNAVAILABLE = 503
HTTP_504_GATEWAY_TIMEOUT = 504
HTTP_505_HTTP_VERSION_NOT_SUPPORTED = 505
HTTP_506_VARIANT_ALSO_NEGOTIATES = 506
HTTP_507_INSUFFICIENT_STORAGE = 507
HTTP_508_LOOP_DETECTED = 508
HTTP_510_NOT_EXTENDED = 510
HTTP_511_NETWORK_AUTHENTICATION_REQUIRED = 511


WS_1000_NORMAL_CLOSURE = 1000
WS_1001_GOING_AWAY = 1001
WS_1002_PROTOCOL_ERROR = 1002
WS_1003_UNSUPPORTED_DATA = 1003
WS_1007_INVALID_FRAME_PAYLOAD_DATA = 1007
WS_1008_POLICY_VIOLATION = 1008
WS_1009_MESSAGE_TOO_BIG = 1009
WS_1010_MANDATORY_EXT = 1010
WS_1011_INTERNAL_ERROR = 1011
WS_1012_SERVICE_RESTART = 1012
WS_1013_TRY_AGAIN_LATER = 1013
WS_1014_BAD_GATEWAY = 1014

"""
You can't use these status codes manually:

WS_RESERVED = 1004
WS_NO_STATUS_RCVD = 1005
WS_ABNORMAL_CLOSURE = 1006
WS_TLS_HANDSHAKE = 1015
"""

status_text = {int(x[5:8]): x[9:].replace('_', ' ').title() for x in globals() if x.startswith('HTTP')}
status_text.update({int(x[3:7]): x[8:].replace('_', ' ').title() for x in globals() if x.startswith('WS')})


def is_informational(status_code: int):
    return bool(100 <= status_code < 200)


def is_success(status_code: int):
    return bool(200 <= status_code < 300)


def is_redirect(status_code: int):
    return bool(300 <= status_code < 400)


def is_client_error(status_code: int):
    return bool(400 <= status_code < 500)


def is_server_error(status_code: int):
    return bool(500 <= status_code < 600)



import importlib.util
import os
import sys

from rich import print as rich_print
from rich.console import Console
from rich.prompt import Prompt

from panther.cli.utils import cli_error
from panther.configs import config
from panther.db.models import BaseUser
from panther.utils import run_coroutine


def get_password(prompt: str) -> str:
    rich_print(f'[b]{prompt}: [/b]', end='', flush=True)
    password = ''

    if sys.platform == 'win32':
        import msvcrt

        while True:
            ch = msvcrt.getch()
            if ch in (b'\r', b'\n'):
                print()
                break
            elif ch == b'\x08':  # Backspace
                if password:
                    password = password[:-1]
                    sys.stdout.write('\b \b')
                    sys.stdout.flush()
            elif ch == b'\x03':  # Ctrl+C
                raise KeyboardInterrupt
            else:
                try:
                    char = ch.decode()
                    password += char
                    sys.stdout.write('*')
                    sys.stdout.flush()
                except UnicodeDecodeError:
                    continue
    else:
        import termios
        import tty

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            while True:
                ch = sys.stdin.read(1)
                if ch in ('\r', '\n'):
                    print()
                    break
                elif ch == '\x7f':  # Backspace
                    if password:
                        password = password[:-1]
                        sys.stdout.write('\b \b')
                        sys.stdout.flush()
                elif ch == '\x03':  # Ctrl+C
                    raise KeyboardInterrupt
                else:
                    password += ch
                    sys.stdout.write('*')
                    sys.stdout.flush()
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return password


def get_username(prompt: str, user_model: BaseUser, username_field: str):
    console = Console(style='bold')
    username = Prompt.ask(prompt=prompt, console=console)
    while run_coroutine(user_model.exists({username_field: username})):
        console.print(f'{user_model.__name__} with this {username_field} already exists.', style='bold red')
        username = Prompt.ask(prompt=prompt, console=console)
    return username


def load_application_file(file_location: str):
    file_name = file_location.removesuffix('.py')
    script_path = os.path.abspath(path=file_location)
    script_dir = os.path.dirname(script_path)
    if script_dir not in sys.path:
        sys.path.insert(0, script_dir)
    spec = importlib.util.spec_from_file_location(name=file_name, location=file_location)
    mod = importlib.util.module_from_spec(spec=spec)
    spec.loader.exec_module(module=mod)


def create_user(args) -> None:
    if len(args) == 0:
        return cli_error(
            'Not Enough Arguments, Give me a file path that contains `Panther()` app.\n'
            '       * Make sure to run `panther createuser` in the same directory as that file!\n'
            '       * Example: `panther createuser main.py`',
        )
    elif len(args) != 1:
        return cli_error('Too Many Arguments.')
    load_application_file(file_location=args[0])
    # Initialization
    User: BaseUser = config.USER_MODEL
    user_model_name = User.__name__
    username_field = User.USERNAME_FIELD
    console = Console()
    try:
        username = get_username(f'Enter your `{username_field}`', user_model=User, username_field=username_field)
        password = get_password('Enter your `password`')
    except KeyboardInterrupt:
        console.print('\nKeyboard Interrupt', style='bold red')
        return

    # Creating User
    user = run_coroutine(User.insert_one({username_field: username}))
    run_coroutine(user.set_password(password=password))
    console.print(f'\n{user_model_name}({username_field}={username}) Created Successfully.', style='bold green')



import os
import shutil
import sys
from io import StringIO
from pathlib import Path
from unittest import skipIf
from unittest.async_case import IsolatedAsyncioTestCase
from unittest.mock import patch

from rich import print as rprint

from panther import Panther
from panther.cli.create_project_command import CreateProject
from panther.cli.create_user_command import create_user
from panther.cli.template import SINGLE_FILE_TEMPLATE, TEMPLATE
from panther.configs import config
from panther.db.connections import db
from panther.db.models import BaseUser

interactive_cli_1_index = 0
interactive_cli_2_index = 0

DB_PATH = 'test_db.pdb'


# 0.ProjectName, 1.BaseDir, 2.IsSingleFile, 3.Database,
# 4.Encryption, 5.Authentication, 6.Monitoring, 7.LogQueries, 8.AutoReformat
def interactive_cli_1_mock_responses(index=None):
    global interactive_cli_1_index
    if index is None:
        index = interactive_cli_1_index
    responses = ['project1', 'project1_dir', 'n', '0', 'y', 'n', 'y', 'y', 'y', 'y']
    response = responses[index]
    interactive_cli_1_index += 1
    return response


def interactive_cli_2_mock_responses(index=None):
    global interactive_cli_2_index
    if index is None:
        index = interactive_cli_2_index
    responses = ['project2', 'project2_dir', 'y', '0', 'y', 'n', 'y', 'y', 'y', 'y']
    response = responses[index]
    interactive_cli_2_index += 1
    return response


class TestCLI(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/sample_project')

    def tearDown(self) -> None:
        if db.is_defined:
            db.session.collection('BaseUser').drop()

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()
        sys.path.pop()

    @skipIf(sys.platform.startswith('win'), 'Not supported in windows')
    async def test_print_info(self):
        with patch('sys.stdout', new=StringIO()) as fake_out1:
            Panther(__name__)

        base_dir = f'{Path(__name__).absolute().parent!s:<41}'
        expected_value = rf"""╭────────────────────────────────────────────────────────────╮
│     ____                 __    __                          │
│    /\  _`\              /\ \__/\ \                         │
│    \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __      │
│     \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\    │
│      \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/     │
│       \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\     │
│        \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/     │
│                                                            │
│   Redis: False                                             │
│   Websocket: False                                         │
│   Monitoring: True                                         │
│   Log Queries: True                                        │
│   Background Tasks: False                                  │
│   Base directory: {base_dir}│
│ * Run "panther monitor" in another session for Monitoring  │
│ * You may want to install `uvloop` for better performance  │
│   `pip install uvloop`                                     │
╰────────────────────────────────────────────────────────────╯"""
        with patch('sys.stdout', new=StringIO()) as fake_out2:
            rprint(expected_value)
        assert fake_out1.getvalue() == fake_out2.getvalue()

    @patch('builtins.input', interactive_cli_1_mock_responses)
    async def test_create_normal_template_with_interactive_cli(self):
        CreateProject().create([])

        project_path = interactive_cli_1_mock_responses(1)
        for file_name, data in SINGLE_FILE_TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()
        shutil.rmtree(project_path)

    @patch('builtins.input', interactive_cli_2_mock_responses)
    async def test_create_single_file_template_with_interactive_cli(self):
        CreateProject().create([])

        project_path = interactive_cli_2_mock_responses(1)
        for file_name, data in SINGLE_FILE_TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()
        shutil.rmtree(project_path)

    async def test_create_on_existence_directory(self):
        project_path = 'test-project-directory'
        os.mkdir(project_path)

        with self.assertLogs(level='ERROR') as captured_error:
            CreateProject().create(['test_project', project_path])

        with self.assertLogs(level='INFO') as captured_info:
            CreateProject().create(['test_project', project_path])

        try:
            assert len(captured_error.records) == 1
            assert captured_info.records[0].getMessage() == f'"{project_path}" directory already exists.'

            assert len(captured_info.records) == 2
            assert captured_info.records[0].getMessage() == f'"{project_path}" directory already exists.'
            assert captured_info.records[1].getMessage() == 'Use "panther -h" for more help'
        except AssertionError:
            raise
        finally:
            os.removedirs(project_path)

    async def test_create_project(self):
        project_path = 'test-project-directory'
        CreateProject().create(['test_project', project_path])

        for file_name, data in TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()

        shutil.rmtree(project_path)

    @patch('panther.cli.create_user_command.get_password', return_value='testpass')
    @patch('panther.cli.create_user_command.get_username', return_value='testuser')
    @patch('panther.cli.create_user_command.load_application_file')
    async def test_create_user_success(self, mock_load_application_file, mock_get_username, mock_get_password):
        global DATABASE
        # Setup
        mock_get_username.side_effect = ['testuser']
        DATABASE = {'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': DB_PATH}}
        Panther(__name__, configs=__name__, urls={})

        # Run
        with patch('sys.stdout', new=StringIO()) as fake_out:
            create_user(['dummy.py'])
        # Check user created
        user = await BaseUser.find_one({'username': 'testuser'})
        assert user is not None
        assert user.check_password('testpass')
        # Cleanup
        await user.delete()
        # Check output
        output = fake_out.getvalue()
        assert 'Created Successfully' in output

    @patch('panther.cli.create_user_command.get_password', return_value='testpass')
    @patch('panther.cli.create_user_command.get_username', return_value='testuser')
    @patch('panther.cli.create_user_command.load_application_file')
    async def test_create_user_duplicate(self, mock_load_application_file, mock_get_username, mock_get_password):
        global DATABASE
        # Setup
        mock_get_username.side_effect = ['testuser', 'testuser2']
        DATABASE = {'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': DB_PATH}}
        Panther(__name__, configs=__name__, urls={})

        # Run 1
        create_user(['dummy.py'])
        assert await BaseUser.exists(username='testuser')
        assert not await BaseUser.exists(username='testuser2')

        # Run 2
        create_user(['dummy.py'])
        assert await BaseUser.exists(username='testuser')
        assert await BaseUser.exists(username='testuser2')

    @patch('panther.cli.create_user_command.get_password', side_effect=KeyboardInterrupt)
    @patch('panther.cli.create_user_command.get_username', return_value='testuser')
    @patch('panther.cli.create_user_command.load_application_file')
    async def test_create_user_keyboard_interrupt(
        self, mock_load_application_file, mock_get_username, mock_get_password
    ):
        global DATABASE
        DATABASE = {'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': DB_PATH}}
        Panther(__name__, configs=__name__, urls={})

        # Run
        with patch('sys.stdout', new=StringIO()) as fake_out:
            create_user(['dummy.py'])
        output = fake_out.getvalue()
        assert 'Keyboard Interrupt' in output



import contextlib
import logging
import os
import platform
import signal
from collections import deque
from pathlib import Path

from rich import box
from rich.align import Align
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table

from panther.cli.utils import import_error
from panther.configs import config
from panther.middlewares.monitoring import WebsocketMonitoringMiddleware

with contextlib.suppress(ImportError):
    from watchfiles import watch

logger = logging.getLogger('panther')


class Monitoring:
    def __init__(self):
        self.rows = deque()
        self.monitoring_log_file = Path(config.BASE_DIR / 'logs' / 'monitoring.log')

    def monitor(self) -> None:
        if error := self.initialize():
            # Don't continue if initialize() has error
            logger.error(error)
            return

        with (
            self.monitoring_log_file.open() as f,
            Live(
                self.generate_table(),
                vertical_overflow='visible',
                screen=True,
            ) as live,
            contextlib.suppress(KeyboardInterrupt),
        ):
            f.readlines()  # Set cursor at the end of the file

            if platform.system() == 'Windows':
                watching = watch(self.monitoring_log_file, force_polling=True)
            else:
                watching = watch(self.monitoring_log_file)

            for _ in watching:
                for line in f.readlines():
                    # line = datetime | method | path | ip:port | response_time(seconds) | status
                    columns = line.split('|')
                    if len(columns) != 2:  # Can be `datetime | ` on initiation.
                        columns[4] = self._clean_response_time(columns[4])
                        self.rows.append(columns)
                        live.update(self.generate_table())

    def initialize(self) -> str:
        # Check requirements
        try:
            from watchfiles import watch
        except ImportError as e:
            return import_error(e, package='watchfiles').args[0]

        # Check log file
        if not self.monitoring_log_file.exists():
            return (
                f'`{self.monitoring_log_file}` file not found. '
                f'Make sure `panther.middlewares.monitoring.MonitoringMiddleware` is in your `MIDDLEWARES`.\n'
                f'documentation: https://PantherPy.github.io/middlewares/#monitoring-middleware'
            )

        # Initialize Deque
        self.update_rows()

        # Register the signal handler
        if platform.system() != 'Windows':
            signal.signal(signal.SIGWINCH, self.update_rows)

    def generate_table(self) -> Panel:
        # 2023-03-24 01:42:52 | GET | /user/317/ | 127.0.0.1:48856 |  0.0366 ms | 200

        table = Table(box=box.MINIMAL_DOUBLE_HEAD)
        table.add_column('Datetime', justify='center', style='magenta', no_wrap=True)
        table.add_column('Method', justify='center', style='cyan', no_wrap=True)
        table.add_column('Path', justify='center', style='cyan', no_wrap=True)
        table.add_column('Client', justify='center', style='cyan')
        table.add_column('Response Time', justify='center', style='blue')
        table.add_column('Status', justify='center', style='blue', no_wrap=True)

        for row in self.rows:
            table.add_row(*row)

        return Panel(
            Align.center(Group(table)),
            box=box.ROUNDED,
            padding=(0, 2),
            title='Monitoring',
            border_style='bright_blue',
        )

    def update_rows(self, *args, **kwargs):
        # Top = -4, Bottom = -2 --> -6
        # Print of each line needs two line, so --> x // 2
        lines = (os.get_terminal_size()[1] - 6) // 2
        self.rows = deque(self.rows, maxlen=lines)

    @classmethod
    def _clean_response_time(cls, response_time: str) -> str:
        if response_time == WebsocketMonitoringMiddleware.ConnectedConnectionTime:
            return response_time
        response_time = float(response_time)
        time_unit = ' s'

        if response_time < 0.01:
            response_time = response_time * 1_000
            time_unit = 'ms'

        elif response_time >= 60:
            response_time = response_time / 60
            time_unit = ' m'

        return f'{round(response_time, 4)} {time_unit}'


monitor = Monitoring().monitor



import typing

from panther.base_websocket import Websocket
from panther.request import Request
from panther.response import Response
from panther.websocket import GenericWebsocket


class HTTPMiddleware:
    """Used only in http requests"""

    def __init__(self, dispatch: typing.Callable):
        self.dispatch = dispatch

    async def __call__(self, request: Request) -> Response:
        return await self.dispatch(request=request)


class WebsocketMiddleware:
    """Used only in ws requests"""

    def __init__(self, dispatch: typing.Callable):
        self.dispatch = dispatch

    async def __call__(self, connection: Websocket) -> GenericWebsocket:
        return await self.dispatch(connection=connection)



import sys
from collections.abc import Iterable, Sequence

from pantherdb import Cursor as PantherDBCursor

from panther.configs import QueryObservable
from panther.db.cursor import Cursor
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import check_connection, log_query
from panther.exceptions import NotFoundAPIError

__all__ = ('Query',)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='Query')


class Query(BaseQuery):
    def __init_subclass__(cls, **kwargs):
        QueryObservable.observe(cls)

    @classmethod
    def _reload_bases(cls, parent):
        if not issubclass(parent, BaseQuery):
            msg = f'Invalid Query Class: `{parent.__name__}` should be subclass of `BaseQuery`'
            raise ValueError(msg)

        if cls.__bases__.count(Query):
            cls.__bases__ = (*cls.__bases__[: cls.__bases__.index(Query) + 1], parent)
        else:
            for kls in cls.__bases__:
                if kls.__bases__.count(Query):
                    kls.__bases__ = (*kls.__bases__[: kls.__bases__.index(Query) + 1], parent)

    # # # # # Find # # # # #
    @classmethod
    @check_connection
    @log_query
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        """
        Get a single document from the database.

        Example:
        -------
            >>> from app.models import User

            >>> await User.find_one(id=1, name='Ali')
            or
            >>> await User.find_one({'id': 1, 'name': 'Ali'})
            or
            >>> await User.find_one({'id': 1}, name='Ali')

        """
        return await super().find_one(_filter, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> PantherDBCursor | Cursor:
        """
        Get documents from the database.

        Example:
        -------
            >>> from app.models import User

            >>> await User.find(age=18, name='Ali')
            or
            >>> await User.find({'age': 18, 'name': 'Ali'})
            or
            >>> await User.find({'age': 18}, name='Ali')

        """
        return await super().find(_filter, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        """
        Get the first document from the database.

        Example:
        -------
            >>> from app.models import User

            >>> await User.first(age=18, name='Ali')
            or
            >>> await User.first({'age': 18, 'name': 'Ali'})
            or
            >>> await User.first({'age': 18}, name='Ali')

        """
        return await super().first(_filter, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        """
        Get the last document from the database.

        Example:
        -------
            >>> from app.models import User

            >>> await User.last(age=18, name='Ali')
            or
            >>> await User.last({'age': 18, 'name': 'Ali'})
            or
            >>> await User.last({'age': 18}, name='Ali')

        """
        return await super().last(_filter, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def aggregate(cls, pipeline: Sequence[dict]) -> Iterable[dict]:
        """
        Perform an aggregation using the aggregation framework on this collection.

        Example:
        -------
            >>> from app.models import User

            >>> pipeline = [
            >>>     {'$match': {...}},
            >>>     {'$unwind': ...},
            >>>     {'$group': {...}},
            >>>     {'$project': {...}},
            >>>     {'$sort': {...}}
            >>>     ...
            >>> ]

            >>> await User.aggregate(pipeline)

        """
        return await super().aggregate(pipeline)

    # # # # # Count # # # # #
    @classmethod
    @check_connection
    @log_query
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        """
        Count the number of documents in this collection.

        Example:
        -------
            >>> from app.models import User

            >>> await User.count(age=18, name='Ali')
            or
            >>> await User.count({'age': 18, 'name': 'Ali'})
            or
            >>> await User.count({'age': 18}, name='Ali')

        """
        return await super().count(_filter, **kwargs)

    # # # # # Insert # # # # #
    @classmethod
    @check_connection
    @log_query
    async def insert_one(cls, _document: dict | None = None, /, **kwargs) -> Self:
        """
        Insert a single document.

        Example:
        -------
            >>> from app.models import User

            >>> await User.insert_one(age=18, name='Ali')
            or
            >>> await User.insert_one({'age': 18, 'name': 'Ali'})
            or
            >>> await User.insert_one({'age': 18}, name='Ali')

        """
        # Step 1: Merge document parameters
        # Combine the _document dict with keyword arguments into a single document
        document = cls._merge(_document, kwargs)

        # Step 2: Process and validate document
        # - Validate data types and structure
        # - Convert Model instances to their IDs for database storage
        # - Handle File objects by saving to disk and storing file paths
        # - Process nested objects and relationships
        final_document = await cls._process_document(document)

        # Step 3: Create model instance
        # - Retrieve Model instances from database using IDs
        # - Open File objects from their stored paths
        # - Build the complete model instance with all relationships
        result = await cls._create_model_instance(document=final_document)

        # Step 4: Insert into database and return
        # - Insert the processed document into the database
        # - Assign the generated ID to the model instance
        # - Return the complete model instance
        result.id = await super().insert_one(final_document)
        return result

    @classmethod
    @check_connection
    @log_query
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        """
        Insert an iterable of documents.

        Example:
        -------
            >>> from app.models import User

            >>> users = [
            >>>     {'age': 18, 'name': 'Ali'},
            >>>     {'age': 17, 'name': 'Saba'},
            >>>     {'age': 16, 'name': 'Amin'}
            >>> ]
            >>> await User.insert_many(users)

        """
        return await super().insert_many(documents)

    # # # # # Delete # # # # #
    @check_connection
    @log_query
    async def delete(self) -> None:
        """
        Delete the document.

        Example:
        -------
            >>> from app.models import User

            >>> user = await User.find_one(name='Ali')

            >>> await user.delete()

        """
        await self.delete_one(id=self.id)

    @classmethod
    @check_connection
    @log_query
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        """
        Delete a single document matching the filter.

        Example:
        -------
            >>> from app.models import User

            >>> await User.delete_one(age=18, name='Ali')
            or
            >>> await User.delete_one({'age': 18, 'name': 'Ali'})
            or
            >>> await User.delete_one({'age': 18}, name='Ali')

        """
        return await super().delete_one(_filter, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        """
        Delete one or more documents matching the filter.

        Example:
        -------
            >>> from app.models import User

            >>> await User.delete_many(age=18, name='Ali')
            or
            >>> await User.delete_many({'age': 18, 'name': 'Ali'})
            or
            >>> await User.delete_many({'age': 18}, name='Ali')

        """
        return await super().delete_many(_filter, **kwargs)

    # # # # # Update # # # # #
    @check_connection
    @log_query
    async def update(self, _update: dict | None = None, /, **kwargs) -> None:
        """
        Update the document.

        Example:
        -------
            >>> from app.models import User

            >>> user = await User.find_one(age=18, name='Ali')

            >>> await user.update(name='Saba', age=19)
            or
            >>> await user.update({'name': 'Saba'}, age=19)
            or
            >>> await user.update({'name': 'Saba', 'age': 19})

        """
        await self.update_one({'id': self.id}, _update, **kwargs)
        await self.reload()

    @classmethod
    @check_connection
    @log_query
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        """
        Update a single document matching the filter.

        Example:
        -------
            >>> from app.models import User

            >>> await User.update_one({'id': 1}, age=18, name='Ali')
            or
            >>> await User.update_one({'id': 1}, {'age': 18, 'name': 'Ali'})
            or
            >>> await User.update_one({'id': 1}, {'age': 18}, name='Ali')

        """
        return await super().update_one(_filter, _update, **kwargs)

    @classmethod
    @check_connection
    @log_query
    async def update_many(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> int:
        """
        Update one or more documents that match the filter.

        Example:
        -------
            >>> from app.models import User

            >>> await User.update_many({'name': 'Saba'}, age=18, name='Ali')
            or
            >>> await User.update_many({'name': 'Saba'}, {'age': 18, 'name': 'Ali'})
            or
            >>> await User.update_many({'name': 'Saba'}, {'age': 18}, name='Ali')

        """
        return await super().update_many(_filter, _update, **kwargs)

    # # # # # Other # # # # #
    @classmethod
    async def all(cls) -> list[Self] | Cursor:
        """
        Alias of find() without args

        Example:
        -------
            >>> from app.models import User

            >>> await User.all()

        """
        return await cls.find()

    @classmethod
    async def find_one_or_insert(cls, _filter: dict | None = None, /, **kwargs) -> tuple[Self, bool]:
        """
        Get a single document from the database.
        or
        Insert a single document.

        Example:
        -------
            >>> from app.models import User

            >>> await User.find_one_or_insert(age=18, name='Ali')
            or
            >>> await User.find_one_or_insert({'age': 18, 'name': 'Ali'})
            or
            >>> await User.find_one_or_insert({'age': 18}, name='Ali')

        """
        if obj := await cls.find_one(_filter, **kwargs):
            return obj, False
        return await cls.insert_one(_filter, **kwargs), True

    @classmethod
    async def find_one_or_raise(cls, _filter: dict | None = None, /, **kwargs) -> Self:
        """
        Example:
        -------
            >>> from app.models import User

            >>> await User.find_one_or_raise(age=18, name='Ali')
            or
            >>> await User.find_one_or_raise({'age': 18, 'name': 'Ali'})
            or
            >>> await User.find_one_or_raise({'age': 18}, name='Ali')

        """
        if obj := await cls.find_one(_filter, **kwargs):
            return obj

        raise NotFoundAPIError(detail=f'{cls.__name__} Does Not Exist')

    @classmethod
    async def exists(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        """
        Check if document exists in collection or not

        Example:
        -------
            >>> from app.models import User

            >>> await User.exists(age=18, name='Ali')
            or
            >>> await User.exists({'age': 18, 'name': 'Ali'})
            or
            >>> await User.exists({'age': 18}, name='Ali')

        """
        return await cls.count(_filter, **kwargs) > 0

    async def save(self) -> None:
        """
        Save the document
            If it has `id` --> Update It
            else --> Insert It

        Example:
        -------
            >>> from app.models import User

            # Update
            >>> user = await User.find_one(name='Ali')
            >>> user.name = 'Saba'
            >>> await user.save()
            or
            # Insert
            >>> user = User(name='Ali')
            >>> await user.save()

        """
        document = self.model_dump(exclude={'id'})
        if self.id:
            await self.update(document)
        else:
            self.id = (await self.insert_one(document)).id

    async def reload(self):
        updated_instance = await self.find_one(id=self.id)
        for field_name in self.__class__.model_fields:
            setattr(self, field_name, getattr(updated_instance, field_name))



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta

from panther.db.connections import redis
from panther.exceptions import ThrottlingAPIError
from panther.request import Request
from panther.utils import round_datetime

# In-memory fallback storage for when Redis is unavailable
_fallback_throttle_storage = defaultdict(int)


@dataclass(repr=False, eq=False)
class Throttle:
    rate: int
    duration: timedelta

    @property
    def time_window(self) -> datetime:
        return round_datetime(datetime.now(), self.duration)

    def build_cache_key(self, request: Request) -> str:
        """
        Generate a unique cache key based on time window, user or IP, and path.
        This method is intended to be overridden by subclasses to customize throttling logic.
        """
        identifier = request.user.id if request.user else request.client.ip
        return f'{self.time_window}-{identifier}-{request.path}'

    async def get_request_count(self, request: Request) -> int:
        """
        Get the current request count for this key from Redis or fallback memory.
        """
        key = self.build_cache_key(request)

        if redis.is_connected:
            value = await redis.get(key)
            return int(value) if value else 0

        return _fallback_throttle_storage.get(key, 0)

    async def increment_request_count(self, request: Request) -> None:
        """
        Increment the request count for this key and ensure TTL is set in Redis.
        """
        key = self.build_cache_key(request)

        if redis.is_connected:
            count = await redis.incrby(key, amount=1)
            if count == 1:
                ttl = int(self.duration.total_seconds())
                await redis.expire(key, ttl)
        else:
            _fallback_throttle_storage[key] += 1

    async def check_and_increment(self, request: Request) -> None:
        """
        Main throttling logic:
        - Raises ThrottlingAPIError if limit exceeded.
        - Otherwise increments the request count.
        """
        count = await self.get_request_count(request)
        remaining = self.rate - count - 1
        reset_time = self.time_window + self.duration
        retry_after = int((reset_time - datetime.now()).total_seconds())

        if remaining < 0:
            raise ThrottlingAPIError(
                headers={
                    'Retry-After': str(retry_after),
                    'X-RateLimit-Reset': str(int(reset_time.timestamp())),
                },
            )

        await self.increment_request_count(request)



from panther import status


class PantherError(Exception):
    pass


class DatabaseError(Exception):
    pass


class BaseError(Exception):
    detail: str | dict | list = 'Internal Server Error'
    status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(self, detail: str | dict | list = None, status_code: int = None, headers: dict = None):
        self.detail = detail or self.detail
        self.status_code = status_code or self.status_code
        self.headers = headers


class APIError(BaseError):
    detail: str | dict | list = 'Internal Server Error'
    status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR


class WebsocketError(BaseError):
    detail: str | dict | list = 'Internal Error'
    status_code: int = status.WS_1011_INTERNAL_ERROR


class RedirectAPIError(APIError):
    def __init__(self, url: str, status_code: int = status.HTTP_302_FOUND):
        super().__init__(headers={'Location': url}, status_code=status_code)


class BadRequestAPIError(APIError):
    detail = 'Bad Request'
    status_code = status.HTTP_400_BAD_REQUEST


class AuthenticationAPIError(APIError):
    detail = 'Authentication Error'
    status_code = status.HTTP_401_UNAUTHORIZED


class AuthorizationAPIError(APIError):
    detail = 'Permission Denied'
    status_code = status.HTTP_403_FORBIDDEN


class NotFoundAPIError(APIError):
    detail = 'Not Found'
    status_code = status.HTTP_404_NOT_FOUND


class MethodNotAllowedAPIError(APIError):
    detail = 'Method Not Allowed'
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED


class UnprocessableEntityError(APIError):
    detail = 'Unprocessable Entity Error'
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY


class UpgradeRequiredError(APIError):
    detail = 'This service requires use of the WebSocket protocol.'
    status_code = status.HTTP_426_UPGRADE_REQUIRED


class ThrottlingAPIError(APIError):
    detail = 'Too Many Request'
    status_code = status.HTTP_429_TOO_MANY_REQUESTS


class InvalidPathVariableAPIError(APIError):
    def __init__(self, value: str, variable_type: type):
        detail = f'Path variable `{value}` is not `{variable_type.__name__}`'
        super().__init__(detail=detail, status_code=status.HTTP_400_BAD_REQUEST)



from panther.app import GenericAPI
from panther.openapi import OutputSchema
from panther.openapi.utils import OpenAPIGenerator
from panther.response import TemplateResponse


class ScalarOpenAPI(GenericAPI):
    output_schema = OutputSchema(exclude_in_docs=True)

    def get(self):
        return TemplateResponse(name='scalar.html', context={'openapi': OpenAPIGenerator.generate_openapi_spec()})


class SwaggerOpenAPI(GenericAPI):
    output_schema = OutputSchema(exclude_in_docs=True)

    def get(self):
        return TemplateResponse(name='swagger.html', context={'openapi': OpenAPIGenerator.generate_openapi_spec()})


class RedocOpenAPI(GenericAPI):
    output_schema = OutputSchema(exclude_in_docs=True)

    def get(self):
        return TemplateResponse(name='redoc.html', context={'openapi': OpenAPIGenerator.generate_openapi_spec()})


class RapiDocOpenAPI(GenericAPI):
    output_schema = OutputSchema(exclude_in_docs=True)

    def get(self):
        return TemplateResponse(name='rapidoc.html', context={'openapi': OpenAPIGenerator.generate_openapi_spec()})


class SpotlightOpenAPI(GenericAPI):
    output_schema = OutputSchema(exclude_in_docs=True)

    def get(self):
        return TemplateResponse(name='spotlight.html', context={'openapi': OpenAPIGenerator.generate_openapi_spec()})



import logging
from collections.abc import Callable
from typing import Literal
from urllib.parse import parse_qsl

import orjson as json
from orjson import JSONDecodeError
from pydantic import ValidationError

from panther._utils import read_multipart_form_data
from panther.base_request import BaseRequest
from panther.exceptions import BadRequestAPIError, UnprocessableEntityError

logger = logging.getLogger('panther')


class Request(BaseRequest):
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        self._data = ...
        self.validated_data = None  # It's been set in API.validate_input()
        super().__init__(scope=scope, receive=receive, send=send)

    @property
    def method(self) -> Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
        return self.scope['method']

    @property
    def data(self) -> dict | bytes:
        """Data before validation"""
        if self._data is ...:
            match (self.headers.content_type or '').split('; boundary='):
                case ['' | 'application/json']:
                    self._data = json.loads(self.__body or b'{}')
                case ['application/x-www-form-urlencoded']:
                    self._data = {k.decode(): v.decode() for k, v in parse_qsl(self.__body)}
                case ['multipart/form-data', boundary]:
                    self._data = read_multipart_form_data(boundary=boundary, body=self.__body)
                case [unknown]:
                    # We don't know the `content-type` so just pass the payload to user
                    logger.warning(f"'{unknown}' Content-Type is not supported")
                    self._data = self.__body
        return self._data

    async def read_body(self) -> None:
        """Read the entire body from an incoming ASGI message."""
        self.__body = b''
        more_body = True
        while more_body:
            message = await self.asgi_receive()
            self.__body += message.get('body', b'')
            more_body = message.get('more_body', False)

    def validate_data(self, model):
        if isinstance(self.data, bytes):
            raise UnprocessableEntityError(detail='Content-Type is not valid')
        if self.data is None:
            raise BadRequestAPIError(detail='Request body is required')
        try:
            # `request` will be ignored in regular `BaseModel`
            self.validated_data = model(**self.data)
        except ValidationError as validation_error:
            error = {'.'.join(str(loc) for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise BadRequestAPIError(detail=error)
        except JSONDecodeError:
            raise UnprocessableEntityError(detail='JSON Decode Error')



from functools import cached_property
from io import BufferedReader, BytesIO
from pathlib import Path

from pydantic import BaseModel, field_validator

from panther import status
from panther.exceptions import APIError
from panther.utils import timezone_now


class File(BaseModel):
    file_name: str
    content_type: str
    file: bytes | None = None
    _file_path: Path | None = None
    _buffer: BytesIO | BufferedReader | None = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.file:
            self._buffer = BytesIO(self.file)
        elif 'file_name' in data:
            self._file_path = Path(data['file_name'])
        self._saved_path: str | None = None

    def __enter__(self):
        if not self._buffer:
            # Open file lazily in binary read mode
            self._buffer = open(self._file_path, 'rb')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._buffer:
            self._buffer.close()
            self._buffer = None

    def read(self, size: int = -1) -> bytes:
        self._ensure_buffer()
        return self._buffer.read(size)

    def seek(self, offset: int, whence: int = 0):
        self._ensure_buffer()
        return self._buffer.seek(offset, whence)

    def tell(self) -> int:
        self._ensure_buffer()
        return self._buffer.tell()

    def write(self, data: bytes):
        if isinstance(self._buffer, BytesIO):
            self._buffer.seek(0, 2)
            self._buffer.write(data)
            self.file = self._buffer.getvalue()  # sync updated bytes
        else:
            raise IOError('Write is only supported for in-memory files.')

    def save(self, path: str | None = None, overwrite: bool = False) -> str:
        # If already saved, return the same path
        if self._saved_path is not None:
            return self._saved_path

        self._ensure_buffer()

        # Handle directory paths (ending with slash)
        if path and str(path).endswith('/'):
            # Treat as directory, use original file name
            base_path = Path(path) / self.file_name
        else:
            base_path = Path(path or self.file_name)

        if not overwrite:
            file_path = base_path
            if file_path.exists():
                # Format: file_YYYYMMDD_HHMMSS[_N].ext
                timestamp = timezone_now().strftime('%Y%m%d_%H%M%S')
                file_path = base_path.with_name(f'{base_path.stem}_{timestamp}{base_path.suffix}')

                # Ensure uniqueness if file with same timestamp exists
                counter = 1
                while file_path.exists():
                    file_path = base_path.with_name(f'{base_path.stem}_{timestamp}_{counter}{base_path.suffix}')
                    counter += 1
        else:
            file_path = base_path

        # Ensure directory exists
        file_path.parent.mkdir(parents=True, exist_ok=True)

        # Write file
        with open(file_path, 'wb') as f:
            f.write(self._buffer.read())
            self._buffer.seek(0)

        # Store the saved path for idempotency
        self._saved_path = str(file_path)
        return self._saved_path

    @cached_property
    def size(self) -> int:
        if self.file is not None:
            return len(self.file)
        if self._file_path:
            return self._file_path.stat().st_size
        return 0

    def _ensure_buffer(self):
        if not self._buffer:
            if self._file_path:
                self._buffer = open(self._file_path, 'rb')
            elif self.file is not None:
                self._buffer = BytesIO(self.file)
            else:
                raise ValueError('No file source to read from.')

    def __repr__(self) -> str:
        return f'{self.__repr_name__()}(file_name={self.file_name}, content_type={self.content_type})'

    __str__ = __repr__


class Image(File):
    @field_validator('content_type')
    @classmethod
    def validate_content_type(cls, content_type: str) -> str:
        if not content_type.startswith('image/'):
            msg = f"{content_type} is not a valid image 'content_type'"
            raise APIError(detail=msg, status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE)
        return content_type



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API, GenericAPI
from panther.authentications import CookieJWTAuthentication, QueryParamJWTAuthentication
from panther.configs import config
from panther.db.models import BaseUser
from panther.request import Request
from panther.test import APIClient


class CustomAuth:
    async def __call__(self, request: Request):
        return 'THIS IS USER'


@API()
async def auth_required_api(request: Request):
    return request.user


@API()
async def refresh_token_api(request: Request):
    return await request.user.refresh_tokens()


@API()
async def logout_api(request: Request):
    return await request.user.logout()


@API(auth=CustomAuth)
async def custom_auth_api(request: Request):
    return request.user


async def custom_auth(req):
    return 'THIS IS USER'


@API(auth=custom_auth)
async def custom_auth_function_api(request: Request):
    return request.user


urls = {
    'auth-required': auth_required_api,
    'refresh-token': refresh_token_api,
    'logout': logout_api,
    'custom-auth': custom_auth_api,
    'custom-auth-function': custom_auth_function_api,
}


class User(BaseUser):
    username: str
    password: str


AUTHENTICATION = 'panther.authentications.JWTAuthentication'
SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
DB_PATH = 'test.pdb'
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': DB_PATH,
    },
}
USER_MODEL = 'tests.test_authentication.User'


class TestJWTAuthentication(IsolatedAsyncioTestCase):
    SHORT_TOKEN = {'Authorization': 'Token TOKEN'}
    NOT_ENOUGH_SEGMENT_TOKEN = {'Authorization': 'Bearer XXX'}
    JUST_BEARER_TOKEN = {'Authorization': 'Bearer'}
    BAD_UNICODE_TOKEN = {'Authorization': 'Bearer علی'}
    BAD_SIGNATURE_TOKEN = {
        'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.JAWUkAU2mWhxcd6MS8r9pd44yBIfkEBmpr3WLeqIccM',
    }
    TOKEN_WITHOUT_USER_ID = {
        'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.PpyXW0PgmGSPaaNirm_Ei4Y2fw9nb4TN26RN1u9RHSo',
    }
    TOKEN = {
        'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxfQ.AF3nsj8IQ6t0ncqIx4quoyPfYaZ-pqUOW4z_euUztPM',
    }

    def setUp(self) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        self.client = APIClient(app=app)

    def tearDown(self) -> None:
        Path(DB_PATH).unlink()
        config.refresh()

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    async def test_user_auth_required_without_auth_class(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = None
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required')
        assert res.status_code == 200
        assert res.data is None
        config.AUTHENTICATION = auth_config

    async def test_user_auth_required_without_token(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required')

        assert res.status_code == 200
        assert res.data is None

    async def test_user_auth_required_with_bad_token_1(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.SHORT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWTAuthentication Error: "Authorization keyword is not valid"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token2(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.JUST_BEARER_TOKEN)

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage() == 'JWTAuthentication Error: "Authorization header must contain 2 parts"'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token3(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.BAD_UNICODE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == (
            "JWTAuthentication Error: \"'latin-1' codec can't encode characters in position 0-2: "
            'ordinal not in range(256)"'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_bad_token4(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.NOT_ENOUGH_SEGMENT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWTAuthentication Error: "Not enough segments"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_invalid_token_signature(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.BAD_SIGNATURE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWTAuthentication Error: "Signature verification failed."'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token_without_user_id(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.TOKEN_WITHOUT_USER_ID)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWTAuthentication Error: "Payload does not have `user_id`"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token_user_not_found(self):
        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', headers=self.TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWTAuthentication Error: "User not found"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_user_auth_required_with_token(self):
        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required', headers={'Authorization': f'Bearer {tokens["access_token"]}'})

        expected_response = {
            'id': user.id,
            'username': 'Username',
            'password': 'Password',
            'last_login': user.last_login.isoformat(),
            'date_created': user.date_created.isoformat(),
        }
        assert res.status_code == 200
        assert res.data == expected_response

    async def test_invalid_refresh_token(self):
        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('refresh-token', headers={'Authorization': f'Bearer {tokens["access_token"]}'})

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage()
            == 'JWTAuthentication Error: "Invalid token type; expected `refresh` token."'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    async def test_refresh_token(self):
        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('refresh-token', headers={'Authorization': f'Bearer {tokens["refresh_token"]}'})

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'Redis is not connected; token revocation is not effective.'

        assert res.status_code == 200
        assert res.data.keys() == {'access_token', 'refresh_token'}

    async def test_cookie_authentication_without_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = CookieJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required', headers={'Authorization': f'Bearer {tokens["access_token"]}'})

        assert res.status_code == 200
        assert res.data is None

        config.AUTHENTICATION = auth_config

    async def test_cookie_authentication_invalid_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = CookieJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('auth-required', {'cookie': f'access_token=s'})

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'CookieJWTAuthentication Error: "Not enough segments"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

        config.AUTHENTICATION = auth_config

    async def test_cookie_authentication(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = CookieJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required', headers={'cookie': f'access_token={tokens["access_token"]}'})

        expected_response = {
            'id': user.id,
            'username': 'Username',
            'password': 'Password',
            'last_login': user.last_login.isoformat(),
            'date_created': user.date_created.isoformat(),
        }
        assert res.status_code == 200
        assert res.data == expected_response

        config.AUTHENTICATION = auth_config

    async def test_cookie_invalid_refresh_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = CookieJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('refresh-token', headers={'cookie': f'access_token={tokens["access_token"]}'})

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage()
            == 'CookieJWTAuthentication Error: "Invalid token type; expected `refresh` token."'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

        config.AUTHENTICATION = auth_config

    async def test_cookie_refresh_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = CookieJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get(
                'refresh-token',
                headers={'cookie': f'access_token={tokens["access_token"]}; refresh_token={tokens["refresh_token"]}'},
            )

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'Redis is not connected; token revocation is not effective.'
        assert res.status_code == 200
        assert res.data.keys() == {'access_token', 'refresh_token'}

        config.AUTHENTICATION = auth_config

    async def test_query_param_authentication_without_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = QueryParamJWTAuthentication

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required')

        assert res.status_code == 200
        assert res.data is None

        config.AUTHENTICATION = auth_config

    async def test_query_param_authentication(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = QueryParamJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('auth-required', query_params={'authorization': tokens['access_token']})

        expected_response = {
            'id': user.id,
            'username': 'Username',
            'password': 'Password',
            'last_login': user.last_login.isoformat(),
            'date_created': user.date_created.isoformat(),
        }
        assert res.status_code == 200
        assert res.data == expected_response

        config.AUTHENTICATION = auth_config

    async def test_query_param_invalid_refresh_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = QueryParamJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('refresh-token', query_params={'authorization': tokens['access_token']})

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage()
            == 'QueryParamJWTAuthentication Error: "Invalid token type; expected `refresh` token."'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

        config.AUTHENTICATION = auth_config

    async def test_query_param_refresh_token(self):
        auth_config = config.AUTHENTICATION
        config.AUTHENTICATION = QueryParamJWTAuthentication

        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('refresh-token', query_params={'authorization': tokens['refresh_token']})

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'Redis is not connected; token revocation is not effective.'
        assert res.status_code == 200
        assert res.data.keys() == {'access_token', 'refresh_token'}

        config.AUTHENTICATION = auth_config

    async def test_custom_auth_class(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('custom-auth')

        assert res.status_code == 200
        assert res.data == 'THIS IS USER'

    async def test_custom_auth_function(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('custom-auth-function')

        assert res.status_code == 200
        assert res.data == 'THIS IS USER'

    async def test_invalid_custom_auth_class(self):
        class SyncCustomAuth:
            def __call__(self, request: Request):
                return 'THIS IS USER'

        try:

            @API(auth=SyncCustomAuth)
            async def invalid_custom_auth_api(request: Request):
                return request.user
        except Exception as e:
            assert e.args[0] == 'SyncCustomAuth.__call__() should be `async`'
        else:
            assert False

    async def test_no_param_custom_auth_class(self):
        class InvalidCustomAuth:
            async def __call__(self):
                return 'THIS IS USER'

        try:

            @API(auth=InvalidCustomAuth)
            async def invalid_custom_auth_api(request: Request):
                return request.user
        except Exception as e:
            assert e.args[0] == 'InvalidCustomAuth.__call__() requires 2 positional argument(s) (self, request).'
        else:
            assert False

    async def test_invalid_custom_auth_function(self):
        def auth(req: str):
            pass

        try:

            @API(auth=auth)
            async def invalid_custom_auth_api(request: Request):
                return request.user
        except Exception as e:
            assert e.args[0] == 'auth() should be `async`'
        else:
            assert False

    async def test_no_param_custom_auth_function(self):
        def auth1():
            pass

        try:

            @API(auth=auth1)
            async def invalid_custom_auth_api(request: Request):
                return request.user
        except Exception as e:
            assert e.args[0] == 'auth1() requires 1 positional argument(s) (request).'
        else:
            assert False

    async def test_class_based_api_invalid_auth_type(self):
        def auth():
            pass

        try:

            class MyAPI(GenericAPI):
                auth = True
        except Exception as e:
            assert e.args[0] == (
                '`bool` is not valid for authentication, it should be a callable, a Class with __call__ '
                'method or a single function.'
            )
        else:
            assert False

    async def test_class_based_api_sync_function_auth(self):
        def auth_func(req):
            pass

        try:

            class MyAPI(GenericAPI):
                auth = auth_func
        except Exception as e:
            assert e.args[0] == 'auth_func() should be `async`'
        else:
            assert False

    async def test_class_based_api_no_param_function_auth(self):
        def auth_func2():
            pass

        try:

            class MyAPI(GenericAPI):
                auth = auth_func2
        except Exception as e:
            assert e.args[0] == 'auth_func2() requires 1 positional argument(s) (request).'
        else:
            assert False

    async def test_class_based_api_sync_class_auth(self):
        class CustomAuth1:
            def __call__(self, req):
                pass

        try:

            class MyAPI(GenericAPI):
                auth = CustomAuth1
        except Exception as e:
            assert e.args[0] == 'CustomAuth1.__call__() should be `async`'
        else:
            assert False

    async def test_class_based_api_no_param_class_auth(self):
        class CustomAuth2:
            def __call__(self):
                pass

        try:

            class MyAPI(GenericAPI):
                auth = CustomAuth2
        except Exception as e:
            assert e.args[0] == 'CustomAuth2.__call__() requires 2 positional argument(s) (self, request).'
        else:
            assert False

    async def test_class_based_api_class_without_call(self):
        class CustomAuth3:
            pass

        try:

            class MyAPI(GenericAPI):
                auth = CustomAuth3
        except Exception as e:
            assert e.args[0] == 'CustomAuth3 must implement __call__() method.'
        else:
            assert False

    async def test_class_based_api_class_with_initiated_auth(self):
        class CustomAuth4:
            pass

        try:

            class MyAPI(GenericAPI):
                auth = CustomAuth4()
        except Exception as e:
            assert (
                e.args[0]
                == '`CustomAuth4` is not valid for authentication, it should be a callable, a Class with __call__ method or a single function.'
            )
        else:
            assert False

    async def test_logout(self):
        user = await User.insert_one(username='Username', password='Password')
        tokens = await user.login()

        with self.assertLogs(level='ERROR') as captured:
            res = await self.client.get('logout', headers={'Authorization': f'Bearer {tokens["access_token"]}'})
        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'Redis is not connected; token revocation is not effective.'
        assert res.status_code == 200

        # TODO: We have to implement a mini redis to work with its functionalities even when its not connected.
        # with self.assertLogs(level='ERROR') as captured:
        #     res = await self.client.get('logout', headers={'Authorization': f'Bearer {tokens["access_token"]}'})
        # assert res.status_code == 401



import sys
from datetime import timedelta
from pathlib import Path
from unittest import TestCase

import tests.sample_project.app.models
from panther import Panther
from panther.configs import Config, config


class TestRun(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/sample_project')

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()
        sys.path.pop()

    def tearDown(self) -> None:
        config.refresh()

    def test_init(self):
        app = Panther(__name__)
        assert isinstance(app, Panther)

    def test_load_configs(self):
        base_dir = Path(__name__).resolve().parent
        secret_key = 'fHrIYx3yK0J_UG0K0zD6miLPNy1esoYXzVsvif6e7rY='
        Panther(__name__)

        assert isinstance(config, Config)
        assert base_dir == config.BASE_DIR
        assert config.LOG_QUERIES is True
        assert config.THROTTLING.rate == 10
        assert config.THROTTLING.duration == timedelta(seconds=10)
        assert secret_key == config.SECRET_KEY
        assert len(config.HTTP_MIDDLEWARES) == 1
        assert len(config.WS_MIDDLEWARES) == 0

        assert config.QUERY_ENGINE.__name__ == 'BasePantherDBQuery'
        assert config.USER_MODEL.__name__ == tests.sample_project.app.models.User.__name__
        assert config.USER_MODEL.__module__.endswith('app.models')
        assert config.JWT_CONFIG.algorithm == 'HS256'
        assert config.JWT_CONFIG.life_time == int(timedelta(days=2).total_seconds())
        assert config.JWT_CONFIG.key == secret_key

        assert '' in config.URLS
        config.URLS.pop('')
        assert 'second' in config.URLS
        config.URLS.pop('second')
        assert config.URLS == {}



import logging
import sys
import types
from importlib import import_module
from multiprocessing import Manager

import jinja2

from panther._utils import check_class_type_endpoint, check_function_type_endpoint, import_class
from panther.background_tasks import _background_tasks
from panther.base_websocket import WebsocketConnections
from panther.cli.utils import import_error
from panther.configs import JWTConfig, config
from panther.db.connections import redis
from panther.db.queries.mongodb_queries import BaseMongoDBQuery
from panther.db.queries.pantherdb_queries import BasePantherDBQuery
from panther.exceptions import PantherError
from panther.middlewares.base import HTTPMiddleware, WebsocketMiddleware
from panther.middlewares.monitoring import MonitoringMiddleware, WebsocketMonitoringMiddleware
from panther.panel.views import HomeView
from panther.routings import finalize_urls, flatten_urls

__all__ = (
    'check_endpoints_inheritance',
    'load_authentication_class',
    'load_auto_reformat',
    'load_background_tasks',
    'load_configs_module',
    'load_database',
    'load_log_queries',
    'load_middlewares',
    'load_other_configs',
    'load_redis',
    'load_secret_key',
    'load_templates_dir',
    'load_throttling',
    'load_timezone',
    'load_urls',
    'load_user_model',
    'load_websocket_connections',
)

logger = logging.getLogger('panther')
monitoring_logger = logging.getLogger('monitoring')


def load_configs_module(module_name: str, /) -> dict:
    """Read the config file as dict"""
    if module_name:
        _module = sys.modules[module_name]
    else:
        try:
            _module = import_module('core.configs')
        except ModuleNotFoundError:
            raise _exception_handler(field='core/configs.py', error='Not Found')
    return _module.__dict__


def load_redis(_configs: dict, /) -> None:
    if redis_config := _configs.get('REDIS'):
        # Check redis module installation
        try:
            from redis.asyncio import Redis
        except ImportError as e:
            raise import_error(e, package='redis')
        redis_class_path = redis_config.get('class', 'panther.db.connections.RedisConnection')
        redis_class = import_class(redis_class_path)
        # We have to create another dict then pop the 'class' else we can't pass the tests
        args = redis_config.copy()
        args.pop('class', None)
        redis_class(**args, init=True)


def load_timezone(_configs: dict, /) -> None:
    if timezone := _configs.get('TIMEZONE'):
        config.TIMEZONE = timezone


def load_templates_dir(_configs: dict, /) -> None:
    if templates_dir := _configs.get('TEMPLATES_DIR', '.'):
        config.TEMPLATES_DIR = templates_dir

    if config.TEMPLATES_DIR == '.':
        config.TEMPLATES_DIR = config.BASE_DIR

    config.JINJA_ENVIRONMENT = jinja2.Environment(
        loader=jinja2.ChoiceLoader(
            loaders=(
                jinja2.FileSystemLoader(searchpath=config.TEMPLATES_DIR),
                jinja2.PackageLoader(package_name='panther', package_path='panel/templates/'),
                jinja2.PackageLoader(package_name='panther', package_path='openapi/templates/'),
            ),
        ),
    )


def load_database(_configs: dict, /) -> None:
    database_config = _configs.get('DATABASE', {})
    if 'engine' in database_config:
        if 'class' not in database_config['engine']:
            raise _exception_handler(field='DATABASE', error='`engine["class"]` not found.')

        engine_class_path = database_config['engine']['class']
        engine_class = import_class(engine_class_path)
        # We have to create another dict then pop the 'class' else we can't pass the tests
        args = database_config['engine'].copy()
        args.pop('class')
        config.DATABASE = engine_class(**args)

        if engine_class_path == 'panther.db.connections.PantherDBConnection':
            config.QUERY_ENGINE = BasePantherDBQuery
        elif engine_class_path == 'panther.db.connections.MongoDBConnection':
            config.QUERY_ENGINE = BaseMongoDBQuery

    if 'query' in database_config:
        if config.QUERY_ENGINE:
            logger.warning('`DATABASE.query` has already been filled.')
        config.QUERY_ENGINE = import_class(database_config['query'])


def load_secret_key(_configs: dict, /) -> None:
    if secret_key := _configs.get('SECRET_KEY'):
        config.SECRET_KEY = secret_key


def load_throttling(_configs: dict, /) -> None:
    if throttling := _configs.get('THROTTLING'):
        config.THROTTLING = throttling


def load_user_model(_configs: dict, /) -> None:
    config.USER_MODEL = import_class(_configs.get('USER_MODEL', 'panther.db.models.BaseUser'))
    if config.USER_MODEL not in config.MODELS:
        config.MODELS.append(config.USER_MODEL)


def load_log_queries(_configs: dict, /) -> None:
    if _configs.get('LOG_QUERIES'):
        config.LOG_QUERIES = True


def load_middlewares(_configs: dict, /) -> None:
    # Collect HTTP Middlewares
    for middleware in _configs.get('MIDDLEWARES') or []:
        # This block is for Backward Compatibility
        if isinstance(middleware, list | tuple):
            if len(middleware) == 1:
                middleware = middleware[0]
            elif len(middleware) == 2:
                _deprecated_warning(
                    field='MIDDLEWARES',
                    message='`data` does not supported in middlewares anymore, as your data is static you may want '
                    'to pass them to your middleware with config variables',
                )
                middleware = middleware[0]
            else:
                raise _exception_handler(
                    field='MIDDLEWARES',
                    error=f'{middleware} should be dotted path or type of a middleware class',
                )

        # `middleware` can be type or path of a class
        if not callable(middleware):
            try:
                middleware = import_class(middleware)
            except (AttributeError, ModuleNotFoundError):
                raise _exception_handler(
                    field='MIDDLEWARES',
                    error=f'{middleware} is not a valid middleware path or type',
                )

        if not issubclass(middleware, HTTPMiddleware):
            raise _exception_handler(field='MIDDLEWARES', error=f'{middleware} is not a sub class of `HTTPMiddleware`')

        if issubclass(middleware, MonitoringMiddleware):
            monitoring_logger.debug('')  # Initiated
            config.MONITORING = True

        config.HTTP_MIDDLEWARES.append(middleware)

    # Collect WebSocket Middlewares
    for middleware in _configs.get('WS_MIDDLEWARES') or []:
        # `middleware` can be type or path of a class
        if not callable(middleware):
            try:
                middleware = import_class(middleware)
            except (AttributeError, ModuleNotFoundError):
                raise _exception_handler(
                    field='WS_MIDDLEWARES',
                    error=f'{middleware} is not a valid middleware path or type',
                )

        if not issubclass(middleware, WebsocketMiddleware):
            raise _exception_handler(
                field='WS_MIDDLEWARES', error=f'{middleware} is not a sub class of `WebsocketMiddleware`'
            )

        if issubclass(middleware, WebsocketMonitoringMiddleware):
            monitoring_logger.debug('')  # Initiated
            config.MONITORING = True

        config.WS_MIDDLEWARES.append(middleware)


def load_auto_reformat(_configs: dict, /) -> None:
    if _configs.get('AUTO_REFORMAT'):
        config.AUTO_REFORMAT = True


def load_background_tasks(_configs: dict, /) -> None:
    if _configs.get('BACKGROUND_TASKS'):
        config.BACKGROUND_TASKS = True
        _background_tasks.initialize()


def load_other_configs(_configs: dict, /) -> None:
    known_configs = set(config.__dataclass_fields__)
    for key, value in _configs.items():
        if key.isupper() and key not in known_configs:
            config[key] = value


def load_urls(_configs: dict, /, urls: dict | None) -> None:
    """
    Return tuple of all urls (as a flat dict) and (as a nested dict)
    """
    if isinstance(urls, dict):
        pass

    elif (url_routing := _configs.get('URLs')) is None:
        raise _exception_handler(field='URLs', error='required.')

    elif isinstance(url_routing, dict):
        error = (
            "can't be 'dict', you may want to pass it's value directly to Panther(). Example: Panther(..., urls=...)"
        )
        raise _exception_handler(field='URLs', error=error)

    elif not isinstance(url_routing, str):
        error = 'should be dotted string.'
        raise _exception_handler(field='URLs', error=error)

    else:
        try:
            urls = import_class(url_routing)
        except ModuleNotFoundError as e:
            raise _exception_handler(field='URLs', error=e)

        if not isinstance(urls, dict):
            raise _exception_handler(field='URLs', error='should point to a dict.')

    config.FLAT_URLS = flatten_urls(urls)
    config.URLS = finalize_urls(config.FLAT_URLS)


def load_authentication_class(_configs: dict, /) -> None:
    """Should be after `load_secret_key()` and `load_urls()`"""
    if authentication := _configs.get('AUTHENTICATION'):
        config.AUTHENTICATION = import_class(authentication)

    if ws_authentication := _configs.get('WS_AUTHENTICATION'):
        config.WS_AUTHENTICATION = import_class(ws_authentication)

    load_jwt_config(_configs)


def load_jwt_config(_configs: dict, /) -> None:
    """Only Collect JWT Config If Authentication Is JWTAuthentication"""
    from panther.authentications import JWTAuthentication

    auth_is_jwt = (config.AUTHENTICATION and issubclass(config.AUTHENTICATION, JWTAuthentication)) or (
        config.WS_AUTHENTICATION and issubclass(config.WS_AUTHENTICATION, JWTAuthentication)
    )
    jwt = _configs.get('JWT_CONFIG', {})
    using_panel_views = HomeView in config.FLAT_URLS.values()
    if auth_is_jwt or using_panel_views:
        if 'key' not in jwt:
            if config.SECRET_KEY is None:
                raise _exception_handler(field='JWTConfig', error='`JWTConfig.key` or `SECRET_KEY` is required.')
            jwt['key'] = config.SECRET_KEY
        config.JWT_CONFIG = JWTConfig(**jwt)


def load_websocket_connections():
    """Should be after `load_redis()`"""
    if config.HAS_WS:
        # Check `websockets`
        try:
            import websockets
        except ImportError as e:
            raise import_error(e, package='websockets')

        # Use the redis pubsub if `redis.is_connected`, else use the `multiprocessing.Manager`
        pubsub_connection = redis.create_connection_for_websocket() if redis.is_connected else Manager()
        config.WEBSOCKET_CONNECTIONS = WebsocketConnections(pubsub_connection=pubsub_connection)


def check_endpoints_inheritance():
    """Should be after `load_urls()`"""
    for endpoint in config.FLAT_URLS.values():
        if endpoint == {}:
            continue

        if isinstance(endpoint, types.FunctionType):
            check_function_type_endpoint(endpoint=endpoint)
        else:
            check_class_type_endpoint(endpoint=endpoint)


def _exception_handler(field: str, error: str | Exception) -> PantherError:
    return PantherError(f"Invalid '{field}': {error}")


def _deprecated_warning(field: str, message: str):
    return logger.warning(f"DEPRECATED '{field}': {message}")



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.db.models import BaseUser
from panther.panel.urls import url_routing
from panther.test import APIClient


class CustomUser(BaseUser):
    username: str
    password: str


class CustomModel(Model):
    name: str
    description: str = 'Default description'
    is_active: bool = True


# Test configuration
AUTHENTICATION = 'panther.authentications.JWTAuthentication'
SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
DB_PATH = 'test_panel.pdb'
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': DB_PATH,
    },
}
USER_MODEL = 'tests.test_panel.CustomUser'


class TestPanel(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls={'': url_routing})
        cls.client = APIClient(app=app)

    def tearDown(self) -> None:
        db.session.collection('CustomUser').drop()
        db.session.collection('CustomModel').drop()

    @classmethod
    def tearDownClass(cls):
        config.refresh()
        Path(DB_PATH).unlink(missing_ok=True)

    @classmethod
    def check_in_cookies(cls, *args, cookies):
        cookies = {cookie[1] for cookie in cookies}
        for arg in args:
            for c in cookies:
                c.startswith(arg)
                break
            else:
                assert False

    async def test_login_view_get(self):
        """Test GET request to login page"""
        res = await self.client.get('login/')
        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_login_view_post_success(self):
        """Test successful login POST request"""
        # Create a test user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')

        # Test successful login
        res = await self.client.post(
            path='login/',
            payload={'username': 'testuser', 'password': 'testpass'},
        )
        self.check_in_cookies(b'access_token', b'refresh_token', cookies=res.cookies)
        assert res.status_code == 302  # Redirect

    async def test_login_view_post_user_not_found(self):
        """Test login POST request with non-existent user"""
        res = await self.client.post(
            path='login/',
            payload={'username': 'nonexistent', 'password': 'testpass'},
        )

        assert res.status_code == 400
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_login_view_post_wrong_password(self):
        """Test login POST request with wrong password"""
        # Create a test user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')

        # Test wrong password
        res = await self.client.post(
            path='login/',
            payload={'username': 'testuser', 'password': 'wrongpass'},
        )

        assert res.status_code == 400
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_login_view_post_with_redirect(self):
        """Test login POST request with redirect_to parameter"""
        # Create a test user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')

        # Test login with redirect
        res = await self.client.post(
            path='login/',
            payload={'username': 'testuser', 'password': 'testpass'},
            query_params={'redirect_to': '/admin/dashboard'},
        )

        assert res.status_code == 302
        self.check_in_cookies(b'access_token', b'refresh_token', cookies=res.cookies)

    async def test_home_view_without_auth(self):
        """Test home view without authentication"""
        res = await self.client.get('')
        assert res.status_code == 302  # Should redirect to login

    async def test_home_view_with_auth(self):
        """Test home view with authentication"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test home view with auth
        res = await self.client.get(path='', headers={'Cookie': f'access_token={tokens["access_token"]}'})

        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_table_view_without_auth(self):
        """Test table view without authentication"""
        res = await self.client.get('0/')
        assert res.status_code == 302  # Should redirect to login

    async def test_table_view_with_auth_empty_data(self):
        """Test table view with authentication but no data"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test table view with auth but no data
        res = await self.client.get(path='0/', headers={'Cookie': f'access_token={tokens["access_token"]}'})

        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_table_view_with_auth_with_data(self):
        """Test table view with authentication and data"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Create some test data
        await CustomModel.insert_one(name='Test Item 1', description='Description 1')
        await CustomModel.insert_one(name='Test Item 2', description='Description 2')

        # Test table view with auth and data
        res = await self.client.get(
            path='1/',  # CustomModel should be at index 1 (after CustomUser at index 0)
            headers={'Cookie': f'access_token={tokens["access_token"]}'},
        )

        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_create_view_get_without_auth(self):
        """Test create view GET without authentication"""
        res = await self.client.get('0/create/')
        assert res.status_code == 302  # Should redirect to login

    async def test_create_view_get_with_auth(self):
        """Test create view GET with authentication"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test create view GET with auth
        res = await self.client.get(path='0/create/', headers={'Cookie': f'access_token={tokens["access_token"]}'})

        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_create_view_post_without_auth(self):
        """Test create view POST without authentication"""
        res = await self.client.post('0/create/', payload={'username': 'newuser', 'password': 'newpass'})
        assert res.status_code == 302  # Should redirect to login

    async def test_create_view_post_with_auth_user_model(self):
        """Test create view POST with authentication for user model"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test create user with auth
        res = await self.client.post(
            path='0/create/',
            payload={'username': 'newuser', 'password': 'newpass'},
            headers={'Cookie': f'access_token={tokens["access_token"]}'},
        )

        assert res.status_code == 200
        assert res.data['username'] == 'newuser'
        # Password should be hashed
        assert res.data['password'] != 'newpass'

    async def test_create_view_post_with_auth_regular_model(self):
        """Test create view POST with authentication for regular model"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test create regular model with auth
        res = await self.client.post(
            path='1/create/',
            payload={'name': 'New Item', 'description': 'New Description', 'is_active': True},
            headers={'Cookie': f'access_token={tokens["access_token"]}'},
        )

        assert res.status_code == 200
        assert res.data['name'] == 'New Item'
        assert res.data['description'] == 'New Description'
        assert res.data['is_active'] is True

    async def test_detail_view_get_without_auth(self):
        """Test detail view GET without authentication"""
        res = await self.client.get('0/test-id/')
        assert res.status_code == 302  # Should redirect to login

    async def test_detail_view_get_with_auth(self):
        """Test detail view GET with authentication"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Create a test item
        test_item = await CustomModel.insert_one(name='Test Item', description='Test Description')

        # Test detail view GET with auth
        res = await self.client.get(
            path=f'1/{test_item.id}/', headers={'Cookie': f'access_token={tokens["access_token"]}'}
        )

        assert res.status_code == 200
        assert res.headers['Content-Type'] == 'text/html; charset=utf-8'

    async def test_detail_view_put_without_auth(self):
        """Test detail view PUT without authentication"""
        res = await self.client.put('0/test-id/', payload={'username': 'updated'})
        assert res.status_code == 302  # Should redirect to login

    async def test_detail_view_put_with_auth(self):
        """Test detail view PUT with authentication"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Create a test item
        test_item = await CustomModel.insert_one(name='Test Item', description='Test Description')

        # Test detail view PUT with auth
        res = await self.client.put(
            path=f'1/{test_item.id}/',
            payload={'name': 'Updated Item', 'description': 'Updated Description'},
            headers={'Cookie': f'access_token={tokens["access_token"]}'},
        )

        assert res.status_code == 200
        assert res.data['name'] == 'Updated Item'
        assert res.data['description'] == 'Updated Description'

    async def test_detail_view_delete_without_auth(self):
        """Test detail view DELETE without authentication"""
        res = await self.client.delete('0/test-id/')
        assert res.status_code == 302  # Should redirect to login

    async def test_detail_view_delete_with_auth(self):
        """Test detail view DELETE with authentication"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Create a test item
        test_item = await CustomModel.insert_one(name='Test Item', description='Test Description')

        # Test detail view DELETE with auth
        res = await self.client.delete(
            path=f'1/{test_item.id}/', headers={'Cookie': f'access_token={tokens["access_token"]}'}
        )

        assert res.status_code == 204  # No content

        # Verify item was deleted
        deleted_item = await CustomModel.find_one(id=test_item.id)
        assert deleted_item is None

    async def test_invalid_model_index(self):
        """Test accessing invalid model index"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test invalid model index
        res = await self.client.get(path='999/', headers={'Cookie': f'access_token={tokens["access_token"]}'})

        assert res.status_code == 500  # Should raise an error for invalid index

    async def test_invalid_document_id(self):
        """Test accessing invalid document ID"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test invalid document ID
        res = await self.client.get(path='0/invalid-id/', headers={'Cookie': f'access_token={tokens["access_token"]}'})

        assert res.status_code == 404  # Should raise an error for invalid document ID

    async def test_middleware_redirect_to_slash(self):
        """Test RedirectToSlashMiddleware functionality"""
        # Create and login user
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')
        tokens = await user.login()

        # Test without trailing slash (should redirect)
        res = await self.client.get(path='login', headers={'Cookie': f'access_token={tokens["access_token"]}'})
        assert res.status_code == 307  # Redirect to add trailing slash

    async def test_cookie_authentication_invalid_token(self):
        """Test cookie authentication with invalid token"""
        res = await self.client.get(path='', headers={'Cookie': 'access_token=invalid_token'})

        assert res.status_code == 302  # Should redirect to login

    async def test_cookie_authentication_expired_token(self):
        """Test cookie authentication with expired token"""
        # Create a user but don't use their token
        user = await CustomUser.insert_one(username='testuser', password='testpass')
        await user.set_password('testpass')

        # Use an expired token
        expired_token = (
            'eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxLCJpYXQiOjE2MzQ1Njc4NzQsImV4cCI6MTYzNDU2Nzg3NX0.invalid_signature'
        )

        res = await self.client.get(path='', headers={'Cookie': f'access_token={expired_token}'})

        assert res.status_code == 302  # Should redirect to login



from datetime import datetime

from panther import version
from panther.utils import generate_secret_key

apis_py = """from datetime import timedelta

from panther import status
from panther.app import API
from panther.response import Response
from panther.utils import timezone_now


# More Info: https://PantherPy.GitHub.io/api/
@API(cache=timedelta(minutes=1))
async def hello_world_api():
    current_time = timezone_now().isoformat()
    return Response(
        data={'message': f'Hello from Panther! 🐾 | {current_time}'},
        status_code=status.HTTP_200_OK
    )
"""

models_py = """\"""
More Info: https://PantherPy.GitHub.io/model/
\"""
from panther.db import Model
"""

serializers_py = """\"""
More Info: https://PantherPy.GitHub.io/serializer/
\"""
from panther.serializer import ModelSerializer
"""

middlewares_py = """\"""
More Info: https://PantherPy.GitHub.io/middlewares/
\"""
"""

app_urls_py = """from app.apis import hello_world_api

url_routing = {
    '/': hello_world_api,
}
"""

configs_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""{IMPORT_PATH}{IMPORT_COMPLETE_LOAD_ENV}
{BASE_DIR}{SECRET_KEY}{DATABASE}{REDIS}{USER_MODEL}{AUTHENTICATION}

# More Info: https://PantherPy.GitHub.io/timezone/
TIMEZONE = 'UTC'

# More Info: https://PantherPy.GitHub.io/urls/
URLs = 'core.urls.url_routing'
""" % datetime.now().date().isoformat()

env = """SECRET_KEY='%s'
""" % generate_secret_key()

main_py = """from panther import Panther

app = Panther(__name__)
"""

urls_py = """from app.urls import url_routing as app_url_routing

url_routing = {
    '/': app_url_routing,
}
"""

git_ignore = """__pycache__/
.venv/
.idea/
logs/

.env
*.pdb
"""

requirements = """panther==%s
""" % version()

TEMPLATE = {
    'app': {
        '__init__.py': '',
        'apis.py': apis_py,
        'models.py': models_py,
        'serializers.py': serializers_py,
        'urls.py': app_urls_py,
    },
    'core': {
        '__init__.py': '',
        'configs.py': configs_py,
        'middlewares.py': middlewares_py,
        'urls.py': urls_py,
    },
    'main.py': main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}

single_main_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""
from datetime import timedelta{IMPORT_PATH}

from panther import Panther, status
from panther.app import API
from panther.response import Response
from panther.utils import {IMPORT_LOAD_ENV}timezone_now
{BASE_DIR}{SECRET_KEY}{DATABASE}{REDIS}{USER_MODEL}{AUTHENTICATION}

# More Info: https://PantherPy.GitHub.io/timezone/
TIMEZONE = 'UTC'


# More Info: https://PantherPy.GitHub.io/api/
@API(cache=timedelta(minutes=1))
async def hello_world_api():
    current_time = timezone_now().isoformat()
    return Response(
        data={'message': f'Hello from Panther! 🐾 | {current_time}'},
        status_code=status.HTTP_200_OK
    )


url_routing = {
    '/': hello_world_api,
}

app = Panther(__name__, configs=__name__, urls=url_routing)
""" % datetime.now().date().isoformat()

SINGLE_FILE_TEMPLATE = {
    'main.py': single_main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}

DATABASE_PANTHERDB_PART = """

# More Info: https://PantherPy.GitHub.io/database/#pantherdb
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': BASE_DIR / 'database.pdb',
    }
}"""

DATABASE_MONGODB_PART = """

# More Info: https://PantherPy.GitHub.io/database/#mongodb
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.MongoDBConnection',
        'host': '127.0.0.1',
        'port': 27017,
        'database': '{PROJECT_NAME}'
    }
}"""

REDIS_PART = """

# More Info: https://PantherPy.GitHub.io/redis/
REDIS = {
    'class': 'panther.db.connections.RedisConnection',
    'host': '127.0.0.1',
    'port': 6379,
}"""

USER_MODEL_PART = """

# More Info: https://PantherPy.GitHub.io/user_model/
USER_MODEL = 'panther.db.models.BaseUser'"""

AUTHENTICATION_PART = """

# More Info: https://PantherPy.GitHub.io/authentications/
AUTHENTICATION = 'panther.authentications.JWTAuthentication'"""

SECRET_KEY_PART = """

env = load_env(BASE_DIR / '.env')

SECRET_KEY = env['SECRET_KEY']"""

BASE_DIR_PART = """
BASE_DIR = Path(__name__).resolve().parent"""

IMPORT_PATH_PART = """
from pathlib import Path"""

IMPORT_LOAD_ENV_PART = """load_env, """

IMPORT_COMPLETE_LOAD_ENV_PART = """
from panther.utils import load_env"""



import re
import sys

from setuptools import setup


def panther_version() -> str:
    with open('panther/__init__.py') as f:
        return re.search('__version__ = [\'"]([^\'"]+)[\'"]', f.read()).group(1)


VERSION = panther_version()
with open('README.md', encoding='utf-8') as file:
    DESCRIPTION = file.read()

INSTALL_REQUIRES = [
    'pantherdb~=2.3.1',
    'orjson~=3.10.18',
    'pydantic~=2.11.7',
    'rich~=14.0.0',
    'uvicorn~=0.35.0',
    'pytz~=2025.2',
    'Jinja2~=3.1.6',
    'simple-ulid~=1.0.0',
]
if sys.version_info <= (3, 12):
    INSTALL_REQUIRES.append('httptools~=0.6.4')

EXTRAS_REQUIRE = {
    'full': [
        'redis==6.2.0',
        'motor~=3.7.1',
        'ipython~=9.4.0',
        'python-jose~=3.5.0',
        'ruff~=0.12.2',
        'websockets~=15.0.1',
        'cryptography~=45.0.5',
        'watchfiles~=1.1.0',
    ],
}

setup(
    name='panther',
    version=VERSION,
    python_requires='>=3.10',
    author='Ali RajabNezhad',
    author_email='alirn76@yahoo.com',
    url='https://github.com/alirn76/panther',
    description='Fast & Friendly, Web Framework For Building Async APIs',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    include_package_data=True,
    license='BSD-3-Clause license',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    entry_points={
        'console_scripts': ['panther=panther.cli.main:start'],
    },
    package_data={
        'panther': ['cli/*', 'panel/templates/*', 'openapi/templates/*'],
    },
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
)



from panther.middlewares.base import HTTPMiddleware, WebsocketMiddleware  # noqa: F401
from panther.middlewares.cors import CORSMiddleware  # noqa: F401
from panther.middlewares.monitoring import MonitoringMiddleware, WebsocketMonitoringMiddleware  # noqa: F401



from panther.authentications import CookieJWTAuthentication
from panther.exceptions import AuthenticationAPIError, RedirectAPIError
from panther.request import Request


class AdminCookieJWTAuthentication(CookieJWTAuthentication):
    async def __call__(self, request: Request):
        try:
            return await super().__call__(request=request)
        except AuthenticationAPIError:
            raise RedirectAPIError(url=f'login?redirect_to={request.path}')



import contextlib
import os
import sys
from datetime import datetime
from typing import Annotated, ClassVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, PlainSerializer, WrapValidator

from panther.configs import config
from panther.db.queries import Query
from panther.utils import URANDOM_SIZE, scrypt, timezone_now

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseUser')


def validate_object_id(value, handler):
    if config.DATABASE.__class__.__name__ != 'MongoDBConnection':
        return str(value)

    if isinstance(value, bson.ObjectId):
        return value

    try:
        return bson.ObjectId(value)
    except Exception as e:
        msg = 'Invalid ObjectId'
        raise ValueError(msg) from e


ID = Annotated[str, WrapValidator(validate_object_id), PlainSerializer(lambda x: str(x), return_type=str)] | None


class Model(PydanticBaseModel, Query):
    def __init_subclass__(cls, **kwargs):
        if cls.__module__ == 'panther.db.models' and cls.__name__ == 'BaseUser':
            return
        config.MODELS.append(cls)

    id: ID = None

    @property
    def _id(self):
        """
        Returns the actual ID value:
            - For MongoDB: returns ObjectId
            - For PantherDB: returns str
        """
        if config.DATABASE.__class__.__name__ == 'MongoDBConnection':
            return bson.ObjectId(self.id)
        return self.id


class BaseUser(Model):
    username: str
    password: str = Field('', max_length=64)
    last_login: datetime | None = None
    date_created: datetime | None = None

    USERNAME_FIELD: ClassVar = 'username'

    @classmethod
    def insert_one(cls, _document: dict | None = None, /, **kwargs) -> Self:
        kwargs['date_created'] = timezone_now()
        return super().insert_one(_document, **kwargs)

    async def login(self) -> dict:
        """Return dict of access and refresh tokens"""
        await self.update(last_login=timezone_now())
        return await config.AUTHENTICATION.login(user=self)

    async def refresh_tokens(self) -> dict:
        """Return dict of new access and refresh tokens"""
        return await config.AUTHENTICATION.refresh(user=self)

    async def logout(self) -> dict:
        return await config.AUTHENTICATION.logout(user=self)

    async def set_password(self, password: str):
        """
        URANDOM_SIZE = 16 char -->
            salt = 16 bytes
            salt.hex() = 32 char
            derived_key = 32 char
        """
        salt = os.urandom(URANDOM_SIZE)
        derived_key = scrypt(password=password, salt=salt, digest=True)

        hashed_password = f'{salt.hex()}{derived_key}'
        await self.update(password=hashed_password)

    def check_password(self, password: str) -> bool:
        size = URANDOM_SIZE * 2
        salt = self.password[:size]
        stored_hash = self.password[size:]
        derived_key = scrypt(password=password, salt=bytes.fromhex(salt), digest=True)

        return derived_key == stored_hash



import asyncio
from collections.abc import Callable
from typing import Literal

import orjson as json

from panther.response import HTMLResponse, PlainTextResponse, Response

__all__ = ('APIClient', 'WebsocketClient')


class RequestClient:
    def __init__(self, app: Callable):
        self.app = app
        self.response = b''

    async def send(self, data: dict):
        if data['type'] == 'http.response.start':
            self.header = data
        else:
            self.response += data['body']

    async def receive(self):
        return {
            'type': 'http.request',
            'body': self.payload,
            'more_body': False,
        }

    async def request(
        self,
        path: str,
        method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS'],
        payload: bytes | dict | None,
        headers: dict,
        query_params: dict,
    ) -> Response:
        headers = [(k.encode(), str(v).encode()) for k, v in headers.items()]
        if not path.startswith('/'):
            path = f'/{path}'

        self.payload = payload if isinstance(payload, bytes) else json.dumps(payload)
        query_params = '&'.join(f'{k}={v}' for k, v in query_params.items())
        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 8585),
            'headers': headers,
            'method': method,
            'path': path,
            'raw_path': path.encode(),
            'query_string': query_params.encode(),
        }
        await self.app(
            scope=scope,
            receive=self.receive,
            send=self.send,
        )
        response_headers = {key.decode(): value.decode() for key, value in self.header['headers']}
        cookies = [(key, value) for key, value in self.header['headers'] if key.decode() == 'Set-Cookie']
        if response_headers['Content-Type'] == 'text/html; charset=utf-8':
            data = self.response.decode()
            response = HTMLResponse(data=data, status_code=self.header['status'], headers=response_headers)

        elif response_headers['Content-Type'] == 'text/plain; charset=utf-8':
            data = self.response.decode()
            response = PlainTextResponse(data=data, status_code=self.header['status'], headers=response_headers)

        elif response_headers['Content-Type'] == 'application/octet-stream':
            data = self.response.decode()
            response = PlainTextResponse(data=data, status_code=self.header['status'], headers=response_headers)

        else:
            data = json.loads(self.response or b'null')
            response = Response(data=data, status_code=self.header['status'], headers=response_headers)
        response.cookies = cookies
        return response


class APIClient:
    def __init__(self, app: Callable):
        self._app = app

    async def _send_request(
        self,
        path: str,
        method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS'],
        payload: dict | None,
        headers: dict,
        query_params: dict,
    ) -> Response:
        request_client = RequestClient(app=self._app)
        return await request_client.request(
            path=path,
            method=method,
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    async def options(
        self,
        path: str,
        headers: dict | None = None,
        query_params: dict | None = None,
    ) -> Response:
        return await self._send_request(
            path=path,
            method='OPTIONS',
            payload=None,
            headers=headers or {},
            query_params=query_params or {},
        )

    async def get(
        self,
        path: str,
        headers: dict | None = None,
        query_params: dict | None = None,
    ) -> Response:
        return await self._send_request(
            path=path,
            method='GET',
            payload=None,
            headers=headers or {},
            query_params=query_params or {},
        )

    async def post(
        self,
        path: str,
        payload: bytes | dict | None = None,
        headers: dict | None = None,
        query_params: dict | None = None,
        content_type: str = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return await self._send_request(
            path=path,
            method='POST',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    async def put(
        self,
        path: str,
        payload: bytes | dict | None = None,
        headers: dict | None = None,
        query_params: dict | None = None,
        content_type: Literal['application/json', 'multipart/form-data'] = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return await self._send_request(
            path=path,
            method='PUT',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    async def patch(
        self,
        path: str,
        payload: bytes | dict | None = None,
        headers: dict | None = None,
        query_params: dict | None = None,
        content_type: Literal['application/json', 'multipart/form-data'] = 'application/json',
    ) -> Response:
        headers = {'content-type': content_type} | (headers or {})
        return await self._send_request(
            path=path,
            method='PATCH',
            payload=payload,
            headers=headers,
            query_params=query_params or {},
        )

    async def delete(
        self,
        path: str,
        headers: dict | None = None,
        query_params: dict | None = None,
    ) -> Response:
        return await self._send_request(
            path=path,
            method='DELETE',
            payload=None,
            headers=headers or {},
            query_params=query_params or {},
        )


class WebsocketClient:
    def __init__(self, app: Callable):
        self.app = app
        self.responses = []

    async def send(self, data: dict):
        self.responses.append(data)

    async def receive(self):
        return {'type': 'websocket.connect'}

    def connect(
        self,
        path: str,
        headers: dict | None = None,
        query_params: dict | None = None,
    ):
        headers = [(k.encode(), str(v).encode()) for k, v in (headers or {}).items()]
        if not path.startswith('/'):
            path = f'/{path}'

        query_params = '&'.join(f'{k}={v}' for k, v in (query_params or {}).items())
        scope = {
            'type': 'websocket',
            'asgi': {'version': '3.0', 'spec_version': '2.3'},
            'http_version': '1.1',
            'scheme': 'ws',
            'server': ('127.0.0.1', 8000),
            'client': ('127.0.0.1', 55330),
            'path': path,
            'raw_path': path.encode(),
            'query_string': query_params.encode(),
            'headers': headers,
            'subprotocols': [],
            'state': {},
        }
        asyncio.run(
            self.app(
                scope=scope,
                receive=self.receive,
                send=self.send,
            ),
        )
        return self.responses



from panther.openapi.views import RapiDocOpenAPI, RedocOpenAPI, ScalarOpenAPI, SpotlightOpenAPI, SwaggerOpenAPI

url_routing = {
    'scalar/': ScalarOpenAPI,
    'swagger/': SwaggerOpenAPI,
    'redoc/': RedocOpenAPI,
    'rapidoc/': RapiDocOpenAPI,
    'spotlight/': SpotlightOpenAPI,
}



from unittest import TestCase

from panther import status


class TestStatus(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.statuses = {
            'HTTP_100_CONTINUE': 100,
            'HTTP_101_SWITCHING_PROTOCOLS': 101,
            'HTTP_102_PROCESSING': 102,
            'HTTP_103_EARLY_HINTS': 103,
            'HTTP_200_OK': 200,
            'HTTP_201_CREATED': 201,
            'HTTP_202_ACCEPTED': 202,
            'HTTP_203_NON_AUTHORITATIVE_INFORMATION': 203,
            'HTTP_204_NO_CONTENT': 204,
            'HTTP_205_RESET_CONTENT': 205,
            'HTTP_206_PARTIAL_CONTENT': 206,
            'HTTP_207_MULTI_STATUS': 207,
            'HTTP_208_ALREADY_REPORTED': 208,
            'HTTP_226_IM_USED': 226,
            'HTTP_300_MULTIPLE_CHOICES': 300,
            'HTTP_301_MOVED_PERMANENTLY': 301,
            'HTTP_302_FOUND': 302,
            'HTTP_303_SEE_OTHER': 303,
            'HTTP_304_NOT_MODIFIED': 304,
            'HTTP_305_USE_PROXY': 305,
            'HTTP_306_RESERVED': 306,
            'HTTP_307_TEMPORARY_REDIRECT': 307,
            'HTTP_308_PERMANENT_REDIRECT': 308,
            'HTTP_400_BAD_REQUEST': 400,
            'HTTP_401_UNAUTHORIZED': 401,
            'HTTP_402_PAYMENT_REQUIRED': 402,
            'HTTP_403_FORBIDDEN': 403,
            'HTTP_404_NOT_FOUND': 404,
            'HTTP_405_METHOD_NOT_ALLOWED': 405,
            'HTTP_406_NOT_ACCEPTABLE': 406,
            'HTTP_407_PROXY_AUTHENTICATION_REQUIRED': 407,
            'HTTP_408_REQUEST_TIMEOUT': 408,
            'HTTP_409_CONFLICT': 409,
            'HTTP_410_GONE': 410,
            'HTTP_411_LENGTH_REQUIRED': 411,
            'HTTP_412_PRECONDITION_FAILED': 412,
            'HTTP_413_REQUEST_ENTITY_TOO_LARGE': 413,
            'HTTP_414_REQUEST_URI_TOO_LONG': 414,
            'HTTP_415_UNSUPPORTED_MEDIA_TYPE': 415,
            'HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE': 416,
            'HTTP_417_EXPECTATION_FAILED': 417,
            'HTTP_418_IM_A_TEAPOT': 418,
            'HTTP_421_MISDIRECTED_REQUEST': 421,
            'HTTP_422_UNPROCESSABLE_ENTITY': 422,
            'HTTP_423_LOCKED': 423,
            'HTTP_424_FAILED_DEPENDENCY': 424,
            'HTTP_425_TOO_EARLY': 425,
            'HTTP_426_UPGRADE_REQUIRED': 426,
            'HTTP_428_PRECONDITION_REQUIRED': 428,
            'HTTP_429_TOO_MANY_REQUESTS': 429,
            'HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE': 431,
            'HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS': 451,
            'HTTP_500_INTERNAL_SERVER_ERROR': 500,
            'HTTP_501_NOT_IMPLEMENTED': 501,
            'HTTP_502_BAD_GATEWAY': 502,
            'HTTP_503_SERVICE_UNAVAILABLE': 503,
            'HTTP_504_GATEWAY_TIMEOUT': 504,
            'HTTP_505_HTTP_VERSION_NOT_SUPPORTED': 505,
            'HTTP_506_VARIANT_ALSO_NEGOTIATES': 506,
            'HTTP_507_INSUFFICIENT_STORAGE': 507,
            'HTTP_508_LOOP_DETECTED': 508,
            'HTTP_510_NOT_EXTENDED': 510,
            'HTTP_511_NETWORK_AUTHENTICATION_REQUIRED': 511,
            'WS_1000_NORMAL_CLOSURE': 1000,
            'WS_1001_GOING_AWAY': 1001,
            'WS_1002_PROTOCOL_ERROR': 1002,
            'WS_1003_UNSUPPORTED_DATA': 1003,
            'WS_1007_INVALID_FRAME_PAYLOAD_DATA': 1007,
            'WS_1008_POLICY_VIOLATION': 1008,
            'WS_1009_MESSAGE_TOO_BIG': 1009,
            'WS_1010_MANDATORY_EXT': 1010,
            'WS_1011_INTERNAL_ERROR': 1011,
            'WS_1012_SERVICE_RESTART': 1012,
            'WS_1013_TRY_AGAIN_LATER': 1013,
            'WS_1014_BAD_GATEWAY': 1014,
        }

    def test_status_numbers(self):
        for _status, number in self.statuses.items():
            assert getattr(status, _status) == number

    def test_status_is_informational(self):
        for number in range(100, 600):
            if 100 <= number < 200:
                assert status.is_informational(number) is True
            else:
                assert status.is_informational(number) is False

    def test_status_is_success(self):
        for number in range(100, 600):
            if 200 <= number < 300:
                assert status.is_success(number) is True
            else:
                assert status.is_success(number) is False

    def test_status_is_redirect(self):
        for number in range(100, 600):
            if 300 <= number < 400:
                assert status.is_redirect(number) is True
            else:
                assert status.is_redirect(number) is False

    def test_status_is_client_error(self):
        for number in range(100, 600):
            if 400 <= number < 500:
                assert status.is_client_error(number) is True
            else:
                assert status.is_client_error(number) is False

    def test_status_is_server_error(self):
        for number in range(100, 600):
            if 500 <= number < 600:
                assert status.is_server_error(number) is True
            else:
                assert status.is_server_error(number) is False

    def test_statuses_text(self):
        for _status, number in self.statuses.items():
            if _status.startswith('HTTP'):
                assert status.status_text[number] == _status[9:].replace('_', ' ').title()
            else:  # WS
                assert status.status_text[number] == _status[8:].replace('_', ' ').title()
